import numpy as np
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.analysis.bond_valence import BVAnalyzer
from ase.data import atomic_numbers, covalent_radii

from .data import bvse_data, principle_number



def assign_oxidation_states(atoms,
                   max_radius=4.0,
                   symm_tol = 1e-3,
                   max_permutations=10000,
                   forbidden_species = ['O1-'],
                   distance_scale_factor=1.05,
                   charge_neutrality_tolerance=1e-5,
                   ):
    
    """
    Decorate ase.Atoms with oxidation states using
    pymatgen's BVAnalyzer.
    
    Parameters 
    ----------

    atoms: ASE's Atoms object
        atomic structure
    
    see BVAnalyzer class in https://pymatgen.org/pymatgen.analysis.html
    for the details regarding other parameters

    Returns
    -------
    atoms with oxidation states stored in atoms.arrays['oxi_states']
    
    """
            
    bva = BVAnalyzer(forbidden_species=forbidden_species,
                    symm_tol=symm_tol,
                    max_radius=max_radius,
                    max_permutations=max_permutations,
                    distance_scale_factor=distance_scale_factor,
                    charge_neutrality_tolerance=charge_neutrality_tolerance
                    )
    st = AseAtomsAdaptor().get_structure(atoms)
    st = bva.get_oxi_state_decorated_structure(st)
    oxi_states = AseAtomsAdaptor.get_atoms(st).arrays['oxi_states']
    atoms.set_array('oxi_states', oxi_states)
    return atoms



def assign_bvse_params(atoms, symbol, charge, self_interaction = True):

    """
    Collect BVSE parameters for a given structure and specific ion
    
    Parameters 
    ----------

    atoms: ASE's Atoms object
        atomic structure
    
    symbol: str
        atomic symbol of the ion of intertest

    charge: int (can be negative)
        formal charge of the ion of interest

    self_interaction: boolean, True by default
        consider interaction between ions of interest (e.g. Li-Li repulsion)
    
    """
    
    s1 = symbol
    z1 = atomic_numbers[symbol]
    q1 = charge
    r0, r_min, alpha, d0, n2, rc2, mask = [], [], [], [], [], [], []
    for _, (s2, q2, z2) in enumerate(zip(atoms.symbols, atoms.get_array('oxi_states'), atoms.numbers)):
        if q2 < 0:
            params = bvse_data[s1][q1][s2][q2]
            r0.append(params['r0']) # dont need it, actually
            r_min.append(params['r_min'])
            alpha.append(params['alpha'])
            d0.append(params['d0'])
            n2.append(principle_number[z2]); rc2.append(covalent_radii[z2])
            mask.append(1.0)
        else:
            r0.append(0); r_min.append(0); alpha.append(0); d0.append(0)
            n2.append(principle_number[z2])
            rc2.append(covalent_radii[z2])
            if z1 == z2:
                if self_interaction:
                    mask.append(1.0)
                else:
                    mask.append(0.0)
            else:
                mask.append(1.0)
    coulomb_params = {'n2': n2, 'rc2': rc2, 'mask': mask}
    morse_params = {'r0': r0, 'd0': d0, 'alpha': alpha, 'r_min': r_min}
    for key in morse_params:
        atoms.set_array(key, np.array(morse_params[key]))
    for key in coulomb_params:
        atoms.set_array(key, np.array(coulomb_params[key]))
    return atoms



import numpy as np
from scipy.spatial import Voronoi
from spglib import get_symmetry_dataset
from ase.data import covalent_radii, chemical_symbols
from ase.spacegroup import get_spacegroup

from ..percolator import Percolator
from ..geometry import _solid_angle, _vol_tetra, _effective_CN, _find_vneighbors, repeated_supercell
from ..data import ionic_radii, elneg_pauling


class StructureFeaturizer:

    def __init__(self, atoms, specie, r_cut = 10.0, oxi = False, symprec = 0.1):
        
        """ 
        Parameters
        ----------

        atoms: ase's Atoms object
            Atomic structure. Should contain a mobile specie of interest.

        specie: int
            Atomic number of a mobile specie, e.g. 3 for Li

        r_cut: float, 10.0 by default
            cutoff for collecting neighbors
            
        oxi: boolean, False by default
            collect features associated with ions

        symprec: float, 0.1 by default
            precision for a space group analysis

        """

        self._set_symmetry_labels(atoms, symprec)
        self.specie = specie
        self.cell = atoms.cell
        self.r_cut = r_cut
        self.atoms = atoms.copy()
        self.oxi = oxi

        
        elneg = []
        for s in atoms.symbols:
            elneg.append(elneg_pauling[s])
        atoms.set_array('elneg', np.array(elneg))
        self.atoms.set_array('elneg', np.array(elneg))        

        if self.oxi:
            ri = []
            for s, o in zip(self.atoms.symbols, self.atoms.arrays['oxi_states']):
                ri.append(ionic_radii[s][o])
            self.atoms.set_array('ri', np.array(ri))  
        self.sg = get_spacegroup(self.atoms)



    def _set_symmetry_labels(self, atoms, symprec):
        
        spglib_cell = (atoms.cell,
                       atoms.get_scaled_positions(),
                       atoms.numbers)
        equivalent_sites = get_symmetry_dataset(spglib_cell, symprec)['equivalent_atoms']
        atoms.set_array('sym_label', np.array(equivalent_sites))
        
            

    def _create_supercell(self, atoms):

        cell = atoms.cell
        scale_a = np.ceil(self.r_cut/(atoms.cell.volume/np.linalg.norm(np.cross(cell[2, :], cell[1, :]))))
        scale_b = np.ceil(self.r_cut/(atoms.cell.volume/np.linalg.norm(np.cross(cell[0, :], cell[2, :]))))
        scale_c = np.ceil(self.r_cut/(atoms.cell.volume/np.linalg.norm(np.cross(cell[0, :], cell[1, :]))))
        scale = np.array([scale_a, scale_b, scale_c])
        scale = np.where(scale < 3, 3, scale)        
        #mobile_atoms = self.atoms[self.atoms.numbers == self.specie]
        #supercell = Box(atoms)._super_box(scale, center = True)
        supercell = repeated_supercell(atoms, scale)
        
        # this update was not tested
        shift = 0.5 * np.dot(np.array(scale) - 1, cell)
        supercell.translate(shift)
        supercell.wrap()
        return supercell



    def _voro_data(self, supercell, central_points_ids):
        """
        Modified pymatgen's version
        """
        
        # if mobile_sublattice:
        #     central_points_ids = central_points_ids[atoms.numbers == self.specie]


        voro = Voronoi(supercell.positions)
        all_vertices = voro.vertices
        positions = supercell.positions
        # Get the coordinates of the central site

        poly_data = {}
        all_vertices = voro.vertices
        positions = voro.points
        unitcell_ids = supercell.get_array('index')
        for site_id in central_points_ids:
            center_coords = positions[site_id]
            data = {
                'volume': [],
                'area': [],
                'face_dist': [],
                'solid_angle': [],
                'weight': [],
                'nn_id': [],
                'nn_id_unitcell': [],
            }
            if self.oxi:
                data.update({'ionic_dist': []})

            #other_sites = []
            for nn, vind in voro.ridge_dict.items():
                # Get only those that include the site in question
                if site_id in nn:
                    other_site = nn[0] if nn[1] == site_id else nn[1]
                    if -1 in vind:
                        raise RuntimeError("This structure is pathological, infinite vertex in the Voronoi construction")
                    # Get the solid angle of the face
                    
                    facets = [all_vertices[i] for i in vind]
                    angle = _solid_angle(center_coords, facets)
                    # Compute the volume of associated with this face
                    volume = 0
                    # qvoronoi returns vertices in CCW order, so I can break
                    # the face up in to segments (0,1,2), (0,2,3), ... to compute
                    # its area where each number is a vertex size
                    for j, k in zip(vind[1:], vind[2:]):
                        volume += _vol_tetra(
                            center_coords,
                            all_vertices[vind[0]],
                            all_vertices[j],
                            all_vertices[k],
                        )
                    # Compute the distance of the site to the face
                    face_dist = np.linalg.norm(center_coords - positions[other_site]) / 2
                    # Compute the area of the face (knowing V=Ad/3)
                    face_area = 3 * volume / face_dist
                    # Compute the normal of the facet
                    normal = np.subtract(positions[other_site], center_coords)
                    normal /= np.linalg.norm(normal)

                    data['nn_id'].append(other_site)
                    data['nn_id_unitcell'].append(unitcell_ids[other_site])
                    data['volume'].append(volume)
                    data['area'].append(face_area)
                    data['solid_angle'].append(angle)
                    data['face_dist'].append(face_dist)   
                    if self.oxi: 
                        ri = supercell.arrays['ri']
                        data['ionic_dist'].append(2 * face_dist - ri[[site_id, other_site]].sum())
                    
            for key in data:
                data[key] = np.array(data[key])
            data['effective_CN'] = _effective_CN(data['area'])
            data['weight'] = data['solid_angle']/data['solid_angle'].sum()
            data['elneg'] = np.take(supercell.arrays['elneg'], data['nn_id'])
            data['bond_ionicity'] = abs(supercell.arrays['elneg'][site_id] - supercell.arrays['elneg'][data['nn_id']])
            data['weighted_elneg'] = data['elneg'] * data['weight']
            if self.oxi:
                data['oxi_state'] = np.take(supercell.arrays['oxi_states'], data['nn_id'])
                data['weighted_oxi_state'] = data['oxi_state'] * data['weight']
            poly_data.update({site_id: data})

        return poly_data



    def _percolation_radii(self):
        
        pl = Percolator(self.atoms, mobile_specie=chemical_symbols[self.specie], upper_bound=self.r_cut)
        self.pl = pl
        radii = {
            'r1d': pl.percolation_threshold(1), 
            'r2d': pl.percolation_threshold(2),
            'r3d': pl.percolation_threshold(3)
            }
        return radii
    
    
    def _covalent_free_space(self):
        
        framework = self.atoms[self.atoms.numbers != self.specie].copy()
        v_cell = framework.cell.volume
        v_atoms = (4/3 * np.pi * (covalent_radii[framework.numbers]) ** 3).sum()
        v_free = (v_cell - v_atoms) / len(framework)
        
        return {'framework_covalent_free_space_fraction': v_free}
    
    
    
    def _covalent_packing_fraction(self):
        
        v = self.atoms.cell.volume
        atoms_volume = (4/3 * np.pi * np.array([covalent_radii[n] for n in self.atoms.numbers]) ** 3).sum()
        fraction = (v - atoms_volume) / v
        return {'covalent_packing_fraction_crystal': fraction}
    
    
    
    def _covalent_packing_fraction_framework(self):
        
        v = self.atoms.cell.volume
        rc = np.array([covalent_radii[n] for n in self.atoms.numbers if n != self.specie])
        atoms_volume = (4/3 * np.pi * (rc ** 3)).sum()
        fraction = (v - atoms_volume) / v
        return {'covalent_packing_fraction_framework': fraction}
    
    
    
    
    def _volume_per_atom(self):
        
        vpa = self.atoms.cell.volume / len(self.atoms)
        return {'vpa': vpa}
    

    def _volume_per_atom_framework(self):

        framework = self.atoms[self.atoms.numbers != self.specie].copy()
        vpa = framework.cell.volume / len(framework)
        return {'vpa_framework': vpa}
    

    
    def _free_volume_per_atom_framework(self):
        
        v = self.atoms.cell.volume
        rc = np.array([covalent_radii[n] for n in self.atoms.numbers if n != self.specie])
        atoms_volume = (4/3 * np.pi * (rc ** 3)).sum()
        fvpa = (v - atoms_volume) / len(rc)
        return {'free_framework_volume_per_atom': fvpa}
    
    
    def _calc_stat(self, arr, stat):

        if stat == 'min':
            return arr.min()
        if stat =='max':
            return arr.max()
        if stat == 'range':
            return arr.max() - arr.min()
        if stat == 'std':
            return arr.std()
        if stat == 'mean':
            return arr.mean()   
        
    def _process_site_poly_data(self, site_poly_data, stats = ['min', 'mean']):
        features = {}
        for key in site_poly_data.keys():
            if key in ['nn_id_unitcell', 'nn_id', 'weight']:
                continue
            if key == 'effective_CN':
                features.update({key: site_poly_data[key]})
                continue
            if key in ('volume', 'area'):
                features.update({key: site_poly_data[key].sum()})
                continue
            if key in ('solid_angle'):
                features.update({f'max_{key}': site_poly_data[key].max()})
                continue
            for stat in stats:
                features.update({f'{stat}_{key}': self._calc_stat(site_poly_data[key], stat)})
        return features


    def _mobile_ion_hops_centroids(self):
        mobile_atoms = self.atoms[self.atoms.numbers == self.specie]
        supercell = self._create_supercell(mobile_atoms)
        #central_points_ids = np.arange(len(mobile_atoms) * supercell.info['translation_id'][(0,0,0)],
                                      #(len(mobile_atoms) * (supercell.info['translation_id'][(0,0,0)] + 1)))
        central_points_ids = np.arange(len(mobile_atoms))
    
        nn = _find_vneighbors(supercell.positions, central_points_ids)
        edges = []
        for source in nn.keys():
            for target in nn[source]:
                if source > target:
                    edge = (target, source)
                else:
                    edge = (source, target)
                if edge not in edges:
                    edges.append(edge)
        edges = np.array(edges)
        centroids = (supercell.positions[edges]).sum(axis = 1)/2
        return centroids
        
    
    def featurize(self, stats = ['mean', 'min']):

        """ 
        Parameters
        ----------

        stats: list of str, ['mean', 'min']
            list of statistics to be calculated for each feature
            allowed statistics are 'mean', 'min', 'max', 'range', 'std'

        Returns
        -------
        dict with feature labels and values
        """


        features = {}
        features.update(self._percolation_radii())
        features.update(self._covalent_free_space())
        features.update(self._covalent_packing_fraction())
        features.update(self._covalent_packing_fraction_framework())
        features.update(self._volume_per_atom())
        features.update(self._volume_per_atom_framework())
        features.update(self._free_volume_per_atom_framework())

        centroids = self._mobile_ion_hops_centroids()


        label = 'mobile_ions_framework'
        atoms = self.atoms.copy()
        supercell = self._create_supercell(atoms)
        #central_points_ids = np.arange(len(atoms) * supercell.info['translation_id'][(0,0,0)],
        #                              (len(atoms) * (supercell.info['translation_id'][(0,0,0)] + 1))) 
        central_points_ids = np.arange(len(atoms))
        central_points_ids = central_points_ids[atoms.numbers == self.specie]

        poly_data = self._voro_data(supercell, central_points_ids)
        site_features = {}
        for site in poly_data:
            site_poly_data = poly_data[site]
            data = self._process_site_poly_data(site_poly_data, ['min', 'mean'])
            for key in data.keys():
                if key in site_features.keys():
                    site_features[key].append(data[key])
                else:
                    site_features.update({key: [data[key]]})
        for stat in stats:
            for key in site_features.keys():
                features.update({f'{label}_{stat}_{key}': self._calc_stat(np.array(site_features[key]), stat)})


        label = 'mobile_sublattice'
        atoms = self.atoms[self.atoms.numbers == self.specie].copy()
        supercell = self._create_supercell(atoms)
        #central_points_ids = np.arange(len(atoms) * supercell.info['translation_id'][(0,0,0)],
        #                              (len(atoms) * (supercell.info['translation_id'][(0,0,0)] + 1))) 
        central_points_ids = np.arange(len(atoms))
        central_points_ids = central_points_ids[atoms.numbers == self.specie]
        #poly_data = self._voro_data(supercell, central_points_ids)
        site_features = {}
        for site in poly_data:
            site_poly_data = poly_data[site]
            data = self._process_site_poly_data(site_poly_data, ['min', 'mean'])
            for key in data.keys():
                if key in site_features.keys():
                    site_features[key].append(data[key])
                else:
                    site_features.update({key: [data[key]]})
        for stat in stats:
            for key in site_features.keys():
                features.update({f'{label}_{stat}_{key}': self._calc_stat(np.array(site_features[key]), stat)})

        label = 'centroid'
        atoms = self.atoms[self.atoms.numbers != self.specie].copy()
        supercell = self._create_supercell(atoms)
        central_points_ids = np.arange(len(supercell), len(supercell) + len(centroids))
        for p in centroids:
            supercell.append(self.specie)
            supercell.positions[-1] = p
        poly_data = self._voro_data(supercell, central_points_ids)
        site_features = {}
        for site in poly_data:
            site_poly_data = poly_data[site]
            data = self._process_site_poly_data(site_poly_data, ['min', 'mean'])
            for key in data.keys():
                if key in site_features.keys():
                    site_features[key].append(data[key])
                else:
                    site_features.update({key: [data[key]]})
        for stat in stats:
            for key in site_features.keys():
                features.update({f'{label}_{stat}_{key}': self._calc_stat(np.array(site_features[key]), stat)})



        if self.oxi:
            label = 'anions'
            atoms = self.atoms[self.atoms.arrays['oxi_states'] < 0].copy()
            supercell = self._create_supercell(atoms)
            #central_points_ids = np.arange(len(atoms) * supercell.info['translation_id'][(0,0,0)],
            #                            (len(atoms) * (supercell.info['translation_id'][(0,0,0)] + 1))) 
            central_points_ids = np.arange(len(atoms))
            poly_data = self._voro_data(supercell, central_points_ids)
            site_features = {}
            for site in poly_data:
                site_poly_data = poly_data[site]
                data = self._process_site_poly_data(site_poly_data, ['min', 'mean'])
                for key in data.keys():
                    if key in site_features.keys():
                        site_features[key].append(data[key])
                    else:
                        site_features.update({key: [data[key]]})
            for stat in stats:
                for key in site_features.keys():
                    features.update({f'{label}_{stat}_{key}': self._calc_stat(np.array(site_features[key]), stat)})


            label = 'cations'
            atoms = self.atoms[self.atoms.arrays['oxi_states'] > 0].copy()
            supercell = self._create_supercell(atoms)
            # central_points_ids = np.arange(len(atoms) * supercell.info['translation_id'][(0,0,0)],
            #                             (len(atoms) * (supercell.info['translation_id'][(0,0,0)] + 1))) 
            central_points_ids = np.arange(len(atoms))
            poly_data = self._voro_data(supercell, central_points_ids)
            site_features = {}
            for site in poly_data:
                site_poly_data = poly_data[site]
                data = self._process_site_poly_data(site_poly_data, ['min', 'mean'])
                for key in data.keys():
                    if key in site_features.keys():
                        site_features[key].append(data[key])
                    else:
                        site_features.update({key: [data[key]]})
            for stat in stats:
                for key in site_features.keys():
                    features.update({f'{label}_{stat}_{key}': self._calc_stat(np.array(site_features[key]), stat)})

            label = 'centroid_anions'
            #centroids = self._mobile_ion_hops_centroids()
            atoms = self.atoms[self.atoms.arrays['oxi_states'] < 0].copy()
            supercell = self._create_supercell(atoms)
            central_points_ids = np.arange(len(supercell), len(supercell) + len(centroids))
            for p in centroids:
                supercell.append(self.specie)
                supercell.positions[-1] = p
            poly_data = self._voro_data(supercell, central_points_ids)
            site_features = {}
            for site in poly_data:
                site_poly_data = poly_data[site]
                data = self._process_site_poly_data(site_poly_data, ['min', 'mean'])
                for key in data.keys():
                    if key in site_features.keys():
                        site_features[key].append(data[key])
                    else:
                        site_features.update({key: [data[key]]})
            for stat in stats:
                for key in site_features.keys():
                    features.update({f'{label}_{stat}_{key}': self._calc_stat(np.array(site_features[key]), stat)})
            label = 'centroid_cations'
            #centroids = self._mobile_ion_hops_centroids()
            atoms = self.atoms[self.atoms.arrays['oxi_states'] > 0].copy()
            supercell = self._create_supercell(atoms)
            central_points_ids = np.arange(len(supercell), len(supercell) + len(centroids))
            for p in centroids:
                supercell.append(self.specie)
                supercell.positions[-1] = p
            poly_data = self._voro_data(supercell, central_points_ids)
            site_features = {}
            for site in poly_data:
                site_poly_data = poly_data[site]
                data = self._process_site_poly_data(site_poly_data, ['min', 'mean'])
                for key in data.keys():
                    if key in site_features.keys():
                        site_features[key].append(data[key])
                    else:
                        site_features.update({key: [data[key]]})
            for stat in stats:
                for key in site_features.keys():
                    features.update({f'{label}_{stat}_{key}': self._calc_stat(np.array(site_features[key]), stat)})
        return features



import json
import pickle
from importlib.resources import files

__version__ = "0.2"


_DATA_DIR = files(__package__)


def _load_pickle(filename):
    with (_DATA_DIR / filename).open("rb") as f:
        return pickle.load(f)


def _load_json(filename):
    with (_DATA_DIR / filename).open("r") as f:
        return json.load(f)



_LOADERS = {
    "shannon_data": lambda: _load_pickle("shannon-radii.pickle"),
    "ionic_radii": lambda: _load_pickle("shannon-data-ionic.pickle"),
    "crystal_radii": lambda: _load_pickle("shannon-data-crystal.pickle"),
    "bv_data": lambda: _load_pickle("bvparams2020-average.pickle"),
    "bvse_data": lambda: _load_pickle("bvse_data.pickle"),
    "principle_number": lambda: _load_pickle("quantum_n.pkl"),
    "elneg_pauling": lambda: _load_json("elneg_data.json"),
}


__all__ = list(_LOADERS.keys())


def __getattr__(name):
    if name not in _LOADERS:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

    value = _LOADERS[name]()
    globals()[name] = value  # cache for future access
    return value


import numpy as np
from .geometry import lineseg_dists, repeated_supercell



class Edge:

    def __init__(self, atoms, source, target, offset):

        """
        Helper class for handling ionic jumps (edges) under PBC.

        Parameters
        ----------
        atoms : ase.Atoms
            Structure containing both endpoints.

        source : int
            Index of the moving atom.

        target : int
            Index of the target site atom.

        offset : array-like of float, shape (3,)
            lattice translation connecting source to target.

        """

        self.atoms = atoms.copy()
        self.source = int(source)
        self.target = int(target)
        self.offset = np.array(offset)
        self.cell = self.atoms.cell

    @property
    def p1(self):
        return self.atoms.positions[self.source]

    @property
    def p2(self):
        return self.atoms.positions[self.target] + self.offset @ self.cell

    @property
    def length(self):
        return np.linalg.norm(self.p2 - self.p1)

    @property
    def centroid(self):
        return 0.5 * (self.p1 + self.p2)


    @property
    def wrapped_target(self):
        """
        Find the index of the atom in the unit cell corresponding
        to the periodic image of `target + offset`.
        """

        atoms = self.atoms
        target = self.target
        offset = self.offset

        cell = atoms.cell

        frac_target = cell.scaled_positions(atoms.positions[target]) + offset
        frac_target -= np.floor(frac_target)  # wrap to [0,1)

        frac_atoms = cell.scaled_positions(atoms.positions)
        frac_atoms -= np.floor(frac_atoms)

        ids = np.where(atoms.numbers == atoms.numbers[target])[0]

        dfrac = frac_atoms[ids] - frac_target
        dfrac -= np.round(dfrac)

        disp = dfrac @ cell
        dist = np.linalg.norm(disp, axis=1)

        i = np.argmin(dist)
        if dist[i] > 1e-6:
            raise RuntimeError(
                f"Wrapped target not found (min dist = {dist[i]:.3e} Angst.)"
            )

        return int(ids[i])



    def project_point(self, p):
        """
        Orthogonal projection of point p onto the edge segment.
        """
        ab = self.p2 - self.p1
        t = np.dot(p - self.p1, ab) / np.dot(ab, ab)
        return self.p1 + t * ab



    def nearest_atom_to_edge(self, exclude_endpoints=True):
        """
        Find nearest atom to the edge.

        Returns
        -------
        idx: int
            Atom index
        
        dist: float
            Distance to edge
        
        proj: ndarray (3,)
            Projection point on edge

        """
        pos = self.atoms.positions
        mask = np.ones(len(pos), dtype=bool)

        if exclude_endpoints:
            mask[[self.source, self.wrapped_target]] = False

        candidates = np.where(mask)[0]
        dists = lineseg_dists(pos[candidates], self.p1, self.p2)

        i = np.argmin(dists)
        idx = candidates[i]

        return idx, dists[i], self.project_point(pos[idx])



    def superedge(self, r_cut):
        """
        Create an edge within a supercell

        Parameters
        ----------

        r_cut: float
            supercell minimum size
        
        Returns
        -------
        edge for the supercell

        """
        cell = self.atoms.cell
        h1 = cell.volume/np.linalg.norm(np.cross(cell[2, :], cell[1, :]))
        h2 = cell.volume/np.linalg.norm(np.cross(cell[0, :], cell[2, :]))
        h3 = cell.volume/np.linalg.norm(np.cross(cell[0, :], cell[1, :]))
        scale = np.ceil(r_cut/np.array([h1, h2, h3]))
        supercell = repeated_supercell(self.atoms, scale)
        return Edge(supercell, self.source, self.target, self.offset/scale)
        


    def interpolate(self, n_images=5, center=True):
        
        """
        Generate interpolated images along the edge.

        Parameters
        ----------
        spacing : float
            Approximate spacing between images (Å)
        center : bool
            Center path in the simulation cell

        Returns
        -------
        images : list of ase.Atoms
        """

        traj = np.linspace(self.p1, self.p2, n_images)
        wrapped_target = self.wrapped_target

        images = []
        for p in traj:

            atoms = self.atoms.copy()
            atoms.positions[self.source] = p
    
            moving = np.zeros(len(atoms), dtype=bool)
            moving[self.source] = True
            atoms.set_array("moving", moving)
            
            # remove wrapped target
            keep = [i for i in range(len(atoms)) if i != wrapped_target]
            atoms = atoms[keep]

            images.append(atoms)

        if center:
            shift = 0.5 * self.cell.sum(axis=0) - self.centroid
            for img in images:
                img.positions += shift
                img.wrap()

        return images



    def __repr__(self):
        return (
            f"Edge(source={self.source}, "
            f"target={self.target}, "
            f"offset={tuple(self.offset)}, "
            f"length={self.length:.2f})"
        )



import numpy as np
from scipy.special import erfc

class BVFF:

    def Morse(r, r_min, d0, alpha):
        
        """ Calculate Morse-type interaction energy.
            Note: interaction is calculated between ions 
                    of opposite sign compared to mobile ion


        Parameters
        ----------

        r: np.array of floats
            distance between mobile ion and framework
            
        r_min: np.array of floats
            minimum energy distance between mobile and framework ions
            
        d0: np.array of floats
            bond breaking parameter
            
        alpha: np.array of floats
            inverse BVS tabulated parameter b (alpha = 1 / b)
        
        Returns
        ----------
        
        np.array
            Morse-type interaction energies
        """
        energy = d0 * ((np.exp( alpha * (r_min - r) ) - 1) ** 2 - 1)
        return energy / 2



    def Coulomb(q1, q2, r, rc1, rc2, n1, n2, f = 0.74):
        
        
        """ Calculate Coulombic interaction energy.
            Note: interaction is calculate between ions 
                    of same sign as mobile ion

        Parameters
        ----------

        q1: float.
            formal charge of mobile speice
        
        q2: np.array of floats
            formal charges of framework ions
            
        R: np.array of floats
            distance between mobile ion and framework
            
        rc1: float
            covalent radii of mobile specie

        rc2: np.array of floats
            covalent radii of framework ions
            
        n1: int
            principle quantum numbers of mobile ion
            
        n2: np.array of integers
            principle quantum numbers of framework ions
            
        f: float, 0.74 by default
            screening factor
        
        Returns
        ----------
        
        np.array
            Coulombic interaction energies
        """

        energy = 14.4 * (q1 * q2 / (n2 * n1) ** (1/2)) * (1 / (r)) * erfc(r / (f * (rc2 + rc1)))
        return energy

    
    def dMorse_dX(x, alpha, r_min, d0, r):
        """ Calculate partial derivative of Morse-type pair potential.
            Note: interaction is calculated between ions 
                    of opposite sign compared to mobile ion


        Parameters
        ----------
        x: np.array of floats
            x (or y or z) coordinate of distance vectors between mobile ion and framework

        alpha: np.array of floats
            inverse BVS tabulated parameter b (alpha = 1 / b)

        r_min: np.array of floats
            minimum energy distance between mobile and framework ions
  
        d0: np.array of floats
            bond breaking parameter

        r: np.array of floats
            distance between mobile ion and framework
            
        Returns
        ----------
        
        np.array
            partial derivative of the pair potential
        """
        dfdx = -2 * alpha * d0 * x * (np.exp(alpha * (r_min - r)) - 1) * np.exp(alpha * (r_min - r)) * (1/r)
        return dfdx/2


    def Morse_pairwise_force(alpha, d0, r_min, rij):
        f_ij = alpha * d0 * (np.exp (alpha * (r_min - rij)) - 1) * np.exp (alpha * (r_min - rij))
        return f_ij


    def Coulomb_pairwise_force(q1, q2, rij, f, rc1, rc2, n1, n2):
        k = 14.4
        term1 = k * q1 * q2 * erfc(rij / (f * (rc1 + rc2))) / ((rij ** 2) * np.sqrt(n1 * n2)) 
        term2 = 2 * k * q1 * q2 * np.exp (-(rij/(f * (rc1 + rc2)))**2) / (np.sqrt(np.pi)*f*(rc1 + rc2)*rij*np.sqrt(n1*n2))
        f_ij = term1 + term2
        return f_ij




    def dCoulomb_dX(x, q1, q2, rc1, rc2, n1, n2, r, f = 0.74):

        """ Calculate partial derivative of Coulomb-type pair potential.
            Note: interaction is calculated between ions 
                    of same sign compared to mobile ion


        Parameters
        ----------
        x: np.array of floats
            x (or y or z) coordinate of distance vectors between mobile ion and framework

        q1: float.
            formal charge of mobile speice
        
        q2: np.array of floats
            formal charges of framework ions
            
        R: np.array of floats
            distance between mobile ion and framework
            
        rc1: float
            covalent radii of mobile specie

        rc2: np.array of floats
            covalent radii of framework ions

        n1: int
            principle quantum numbers of mobile ion
            
        n2: np.array of integers
            principle quantum numbers of framework ions
            
        f: float, 0.74 by default
            screening factor
            
        Returns
        ----------
        
        np.array
            partial derivative of the pair potential
        """


        frc = f * (rc1 + rc2)
        A = 14.4 * (q1 * q2 / ((n2 * n1) ** (1/2)))
        z = r/frc
        term1 = - x * erfc(r / frc) / (r**3) # (x^2 + y^2 + z^2) == r^2
        term2 = - x * 2 * np.exp(-z**2) / (r**2 * np.sqrt(np.pi) * frc)
        dfdx =  A * (term1 + term2)
        return dfdx


import numpy as np 
from scipy.spatial import Voronoi
from scipy.spatial.distance import cdist, euclidean
from ase.neighborlist import NeighborList



def repeated_supercell(atoms, scale, center = False, tol=1e-10):

    atoms = atoms.copy()
    atoms.set_array('index', np.arange(len(atoms)))
    supercell = atoms.repeat([int(s) for s in scale])

    if center:
        shift = np.dot([0.5, 0.5, 0.5], supercell.cell)
        supercell.translate(shift)

    frac = np.linalg.solve(atoms.cell.T, supercell.positions.T).T

    # Separate integer and fractional parts
    offsets = np.floor(frac + tol).astype(int)
    # intra_frac = frac - offsets # for debug
    supercell.set_array('offset', offsets)
    return supercell



def lineseg_dists(p, a, b):
    
    """Distance from p to the linear sigment (a, b)"""
    # source:    https://stackoverflow.com/questions/54442057/
    # calculate-the-euclidian-distance-between-an-array-of-points-to-a-line-segment-in/
    # 54442561#54442561 
    
    if np.all(a == b):
        return np.linalg.norm(p - a, axis=1)
    d = np.divide(b - a, np.linalg.norm(b - a))
    s = np.dot(a - p, d)
    t = np.dot(p - b, d)
    h = np.maximum.reduce([s, t, np.zeros(len(p))])
    c = np.cross(p - a, d)
    return np.hypot(h, np.linalg.norm(c, axis = 1))



def project_point_on_edge(p, p1, p2):
    """Orthogonal projection of point p onto the linear segment (p1,p2)"""
    ap = p-p1
    ab = p2 - p1
    result = p1 + np.dot(ap,ab)/np.dot(ab,ab) * ab
    return result


"""some geometry utils copied from pymatgen"""

def _effective_CN(areas):
    CN = (areas.sum()**2) / ((areas**2).sum())
    return CN


def _find_vneighbors(points, central_points_ids, key=1, min_dist=-float("inf"), max_dist=float("inf")):
    """soure: https://github.com/mcs-cice/IonExplorer2/blob/main/geometry.py
     Parameter mod can takes values 1, 2, or 3 that correspond to the
    search for domains adjacent by vertices, edges or faces.
    """

    neighbors = {i: None for i in central_points_ids}
    vor = Voronoi(points)
    for i in central_points_ids:
        cp = points[i]
        region = vor.regions[vor.point_region[i]]
        if -1 in region:
            raise ValueError("The domain for \"" + str(i) + "\" point is not closed!")
        local_neighbors = []
        for j in range(len(points)):
            numb_common_vertices = len(np.intersect1d(region, vor.regions[vor.point_region[j]]))
            if i != j and numb_common_vertices >= key and min_dist < np.linalg.norm(cp - points[j]) < max_dist:
                local_neighbors.append(j)
        neighbors[i] =  local_neighbors
    return neighbors



def _vol_tetra(vt1, vt2, vt3, vt4):
    """
    Copied from pymatgen's repo
    Calculate the volume of a tetrahedron, given the four vertices of vt1,
    vt2, vt3 and vt4.

    Args:
        vt1 (array-like): coordinates of vertex 1.
        vt2 (array-like): coordinates of vertex 2.
        vt3 (array-like): coordinates of vertex 3.
        vt4 (array-like): coordinates of vertex 4.

    Returns:
        (float): volume of the tetrahedron.

    """
    vol_tetra = np.abs(np.dot((vt1 - vt4), np.cross((vt2 - vt4), (vt3 - vt4)))) / 6    
    return vol_tetra



def _solid_angle(center, coords):
    """
    Helper method to calculate the solid angle of a set of coords from the
    center.

    Args:
        center (3x1 array): Center to measure solid angle from.
        coords (Nx3 array): List of coords to determine solid angle.

    Returns:
        The solid angle.
    """
    # Compute the displacement from the center
    r = [np.subtract(c, center) for c in coords]

    # Compute the magnitude of each vector
    r_norm = [np.linalg.norm(i) for i in r]

    # Compute the solid angle for each tetrahedron that makes up the facet
    #  Following: https://en.wikipedia.org/wiki/Solid_angle#Tetrahedron
    angle = 0
    for i in range(1, len(r) - 1):
        j = i + 1
        tp = np.abs(np.dot(r[0], np.cross(r[i], r[j])))
        de = (
            r_norm[0] * r_norm[i] * r_norm[j]
            + r_norm[j] * np.dot(r[0], r[i])
            + r_norm[i] * np.dot(r[0], r[j])
            + r_norm[0] * np.dot(r[i], r[j])
        )
        my_angle = (0.5 * np.pi if tp > 0 else -0.5 * np.pi) if de == 0 else np.arctan(tp / de)
        angle += (my_angle if my_angle > 0 else my_angle + np.pi) * 2
    return angle



def _geometric_median(X, y0, eps=1e-5, max_iter = 200):
    """
    Finds geometric median of the X points using y0 as the initial guess
    """
    #y = np.mean(X, 0)
    y = y0
    counter = 0
    while True:
        counter += 1
        D = cdist(X, [y])
        nonzeros = (D != 0)[:, 0]

        Dinv = 1 / D[nonzeros]
        Dinvs = np.sum(Dinv)
        W = Dinv / Dinvs
        T = np.sum(W * X[nonzeros], 0)

        num_zeros = len(X) - np.sum(nonzeros)
        if num_zeros == 0:
            y1 = T
        elif num_zeros == len(X):
            return y
        else:
            R = (T - y) * Dinvs
            r = np.linalg.norm(R)
            rinv = 0 if r == 0 else num_zeros/r
            y1 = max(0, 1-rinv)*T + min(1, rinv)*y

        if euclidean(y, y1) < eps:
            return y1
        
        if counter > max_iter:
            return y
        y = y1



class VoroSite:

    def __init__(self, atoms, upper_bound = 10.0):
        self.atoms = atoms.copy()
        self.cell = atoms.cell
        self.r_cut = min(upper_bound, self._vcutoff())

    def _vcutoff(self):
        """from pymatgen.analysis.local_env"""
        corners = [[1, 1, 1], [-1, 1, 1], [1, -1, 1], [1, 1, -1]]
        d_corners = [np.linalg.norm(self.atoms.cell.cartesian_positions(c)) for c in corners]
        max_cutoff = max(d_corners) + 0.01
        return max_cutoff
    
    def _vneighbors(self, site_id):
        scale = np.ceil(self.r_cut/abs(np.diag(self.cell)))
        #supercell = Box(self.atoms.copy())._super_box(scale)
        supercell = repeated_supercell(self.atoms, scale)
        shift = np.dot([0.5, 0.5, 0.5], supercell.cell) - supercell.positions[site_id]
        supercell.positions += shift
        supercell.wrap()
        voro = Voronoi(supercell.positions)
        unitcell_ids = supercell.get_array('index')
        return voro, supercell
    
    def _vneighbors2(self, site_id):

        nl = NeighborList([self.r_cut] * len(self.atoms), self_interaction=False)
        nl.update(self.atoms)
        indices, offsets = nl.get_neighbors(site_id)
        coords = self.atoms.positions[indices] + np.dot(offsets, self.cell)
        voro = Voronoi(coords)
        return voro, indices
    
    def get_poly_data(self, site_id):
        voro, supercell = self._vneighbors(site_id)
        all_vertices = voro.vertices
        positions = voro.points
        center_coords = positions[site_id]
        
        data = {
            'volume': [],
            'area': [],
            'face_dist': [],
            'solid_angle': [],
            'weight': [],
            'nn_id': [],
            'nn_id_unitcell': [],
        }
        unitcell_ids = supercell.get_array('index')
        other_sites = []
        for nn, vind in voro.ridge_dict.items():
            # Get only those that include the site in question
            if site_id in nn:
                other_site = nn[0] if nn[1] == site_id else nn[1]
                if -1 in vind:
                    raise RuntimeError("This structure is pathological, infinite vertex in the Voronoi construction")
                # Get the solid angle of the face
                
                facets = [all_vertices[i] for i in vind]
                angle = _solid_angle(center_coords, facets)
                # Compute the volume of associated with this face
                volume = 0
                # qvoronoi returns vertices in CCW order, so I can break
                # the face up in to segments (0,1,2), (0,2,3), ... to compute
                # its area where each number is a vertex size
                for j, k in zip(vind[1:], vind[2:]):
                    volume += _vol_tetra(
                        center_coords,
                        all_vertices[vind[0]],
                        all_vertices[j],
                        all_vertices[k],
                    )
                # Compute the distance of the site to the face
                face_dist = np.linalg.norm(center_coords - positions[other_site]) / 2
                # Compute the area of the face (knowing V=Ad/3)
                face_area = 3 * volume / face_dist
                # Compute the normal of the facet
                normal = np.subtract(positions[other_site], center_coords)
                normal /= np.linalg.norm(normal)
        
                data['nn_id'].append(other_site)
                data['nn_id_unitcell'].append(unitcell_ids[other_site])
                data['volume'].append(volume)
                data['area'].append(face_area)
                data['solid_angle'].append(angle)
                data['face_dist'].append(face_dist)

        for key in data.keys():
            data[key] = np.array(data[key])
        return data, supercell    





import numpy as np
from ase.data import covalent_radii
from ase.stress import full_3x3_to_voigt_6_stress
from ase.neighborlist import NeighborList
from ase.calculators.calculator import Calculator, all_changes

from ..data import principle_number
from .potential import BVFF



class BVSECalculator(Calculator):
    """
    BVSE calculator based on ASE's Calculator. It is used for NEB calculations
    of a mobile ion migration barrier.
    """

    implemented_properties = ["energy", "forces", "stressses", "stress"]
    default_parameters = {
        'rc': 10.0,
    }
    nolabel = True # check

    def __init__(self, sites = None, nl = None, f = 0.74, **kwargs):
        """
        Parameters
        ----------
        sites: list of int
            indices of sites of interest in ase.Atoms
            Note: Considering source, taret notation it is the source
        
        nl: ase NeighborList, optional
            object should have method get_neighbors(site)
        
        """
        super().__init__(**kwargs)
        self.sites = sites
        self.nl = nl
        self.f = f

    def calculate(
        self,
        atoms = None,
        properties = None,
        system_changes = None,
    ):
        """
        Parameters
        ----------

        atoms: ASE's Atoms object
            atomic structure

        properties: list
            list of properties to calculate
        
        system_changes: list
            monitor which properties of atoms were
        
        changed for new calculation. If not, the previous calculation
                results will be loaded.
        Returns:

        """
        properties = properties or ["energy", "forces", "stress"]
        system_changes = system_changes or all_changes
        super().calculate(atoms=atoms, properties=properties, system_changes=system_changes)
        
        natoms = len(self.atoms)
        rc = self.parameters.rc

        if self.nl is None or 'numbers' in system_changes:
            self.nl = NeighborList(
                [rc / 2] * natoms, self_interaction=False, bothways=True, skin = 0.2
            )

        self.nl.update(self.atoms)
        positions = self.atoms.positions
        cell = self.atoms.cell

        energies = np.zeros(natoms)
        forces = np.zeros((natoms, 3))
        stresses = np.zeros((natoms, 3, 3))
        atoms = self.atoms.copy()
        for ii in self.sites:
            
            neighbors, offsets = self.nl.get_neighbors(ii)
            cells = np.dot(offsets, cell)
            distance_vectors = positions[neighbors] + cells - positions[ii]
            r2 = np.linalg.norm(distance_vectors, axis = 1)
            q1 = atoms.get_array('oxi_states')[ii]
            z1 = atoms.numbers[ii]
            n1 = principle_number[z1]
            rc1 = covalent_radii[z1]
            q2 = np.take(atoms.get_array('oxi_states'), neighbors)
            q2 = np.where(q2 * q1 < 0.0, 0, q2)
            n2 = np.take(atoms.get_array('n2'), neighbors)
            rc2 = np.take(atoms.get_array('rc2'), neighbors)
            alpha =  np.take(atoms.get_array('alpha'), neighbors)
            r_min =  np.take(atoms.get_array('r_min'), neighbors)
            d0 =  np.take(atoms.get_array('d0'), neighbors)
            mask = np.take(atoms.get_array('mask'), neighbors)
            e1 = BVFF.Morse(r2, r_min, d0, alpha).sum() 
            e2 = (mask * BVFF.Coulomb(q1, q2, r2, rc1, rc2, n1, n2, f = self.f)).sum()
            e = e1 + e2
            
            pairwise_energies = e
            # alpha * d0 (np.exp (alpha * (r_min - rij)) - 1) * np.exp (alpha * (r_min - rij)), where rij = r2 
            pairwise_forces = BVFF.Morse_pairwise_force(alpha, d0, r_min, r2) #
            # q1, q2, rij, f, rc1, rc2, n1, n2
            pairwise_forces += BVFF.Coulomb_pairwise_force(q1, q2, r2, self.f, rc1, rc2, n1, n2)
            pairwise_forces = pairwise_forces[:, np.newaxis] * -distance_vectors
            energies[ii] += pairwise_energies.sum()
            forces[ii] += pairwise_forces.sum(axis=0)
            stresses[ii] += 0.5 * np.dot(
                pairwise_forces.T, distance_vectors
            )  # equivalent to outer product

        # no lattice, no stress
        if self.atoms.cell.rank == 3:
            stresses = full_3x3_to_voigt_6_stress(stresses)
            self.results['stress'] = stresses.sum(
                axis=0) / self.atoms.get_volume()
            self.results['stresses'] = stresses / self.atoms.get_volume()
        else:
            print('stresses cannot be computed')
        energy = energies.sum()
        self.results['energy'] = energy
        self.results['energies'] = energies
        self.results['free_energy'] = energy
        self.results['forces'] = forces


import numpy as np
from collections import deque



class PeriodicGraph:

    def __init__(self, sources, targets, offsets):
        self.sources = sources
        self.targets = targets
        self.offsets = offsets

    def find_periodicity(self):
        """
        Returns:
        -------
        periodicity: int
            Percolation dimensionality (0, 1, 2, or 3)
        
        translations: list
            independent translation vectors
        """

        # visited[atom_index] = lattice_offset at first encounter
        visited = {}
        translations = []

        all_nodes = np.unique(self.sources) # need to consider sym_labels in future
        
        for start in all_nodes:
            if start in visited:
                continue

            # BFS queue: (atom_index, lattice_offset)
            queue = deque()
            start = int(np.min(self.sources))
            queue.append((start, np.zeros(3, dtype=int)))
            visited[start] = np.zeros(3, dtype=int)

            while queue:
                i, offset_i = queue.popleft()
                # find neighbors of i
                mask = self.sources == i
                for j, off_ij in zip(self.targets[mask], self.offsets[mask]):
                    offset_j = offset_i + off_ij

                    if j not in visited:
                        visited[j] = offset_j
                        queue.append((j, offset_j))
                    else:
                        # found a translational equivalence
                        delta = offset_j - visited[j]
                        if not np.all(delta == 0):
                            translations.append(delta)

        independent = self._independent_vectors(translations)
        return len(independent), independent

    @staticmethod
    def _independent_vectors(vectors, tol=1e-8):
        """
        Returns a list of linearly independent integer translation vectors.
        """
        if not vectors:
            return []

        basis = []
        for v in vectors:
            v = np.array(v, dtype=float)
            if not basis:
                basis.append(v)
                continue

            mat = np.vstack(basis + [v])
            if np.linalg.matrix_rank(mat, tol=tol) > len(basis):
                basis.append(v)

            if len(basis) == 3:
                break

        # convert back to integer vectors
        return [np.round(v).astype(int) for v in basis]


import itertools
import numpy as np
from spglib import get_symmetry_dataset
from ase.io import read
from ase.neighborlist import neighbor_list

from .geometry import lineseg_dists
from .periodic_graph import PeriodicGraph



class Percolator:

    def __init__(self, atoms, mobile_specie=None, upper_bound=8.0, symprec=1e-3):

        """ 
        Parameters
        ----------

        atoms: ase's Atoms object
            Atomic structure. Should contain a mobile specie of interest.

        mobile_specie: str
            Atomic symbol of a mobile specie, e.g. "Li"

        upper_bound: float, 10.0 by default
            maximum jump distance between equilibrium sites of a mobile specie
            
        symprec: float, 1e-3 by default
            precision for a space group analysis
        """

        self.atoms = atoms.copy()
        self.mobile_specie = mobile_specie 
        if mobile_specie not in atoms.symbols:
            raise ValueError(f'No provided mobile_specie ({mobile_specie}) in a structure')
        self.symprec = symprec
        self.upper_bound = upper_bound
        self.edges = None           
        self.edge_lengths = None
        self.edge_min_dists = None

    @classmethod
    def from_file(cls, file, mobile_specie=None, upper_bound=8.0, symprec=1e-3):
        atoms = read(file)
        return cls(atoms, mobile_specie=mobile_specie, upper_bound=upper_bound, symprec=symprec)
    


    def _collect_edges(self):
        
        atoms = self.atoms.copy()
        atoms.set_array('index', np.arange(0, len(atoms)))        
        mobile_atoms = atoms[atoms.symbols == self.mobile_specie].copy()
        sources, targets, offsets, lengths =  neighbor_list('ijSd', mobile_atoms, self.upper_bound)
        # renumerate in accordance with the initial indices in the unitcell
        sources = np.take(mobile_atoms.arrays['index'], sources)
        targets = np.take(mobile_atoms.arrays['index'], targets)
        self.edges = np.column_stack((sources, targets, offsets))
        self.edge_lengths = lengths
        if len(self.edges) == 0:
            print(f'No edges found for the given cutoff={self.upper_bound} and mobile_specie={self.mobile_specie}.')
            raise



    def _annotate_edges(self):

        atoms = self.atoms.copy()
        min_distances = []
        for _, edge in enumerate(self.edges):

            source, target, offset = edge[0], edge[1], edge[2:]
            shift = [0, 0, 0]
            p1 = atoms.positions[source] + np.dot(shift, atoms.cell)
            p2 = atoms.positions[target] + np.dot(offset + shift, atoms.cell)
            base = atoms[[i for i in range(len(atoms)) if i not in [source, target]]]
            translations = np.array(list(itertools.product(
                                            [0, 1, -1], # should be [0, 1, -1, 2, -2] ideally
                                            [0, 1, -1],
                                            [0, 1, -1])))
            #assert abs(np.linalg.norm(p2 - p1) - d_) < 1e-10
            coords = []
            idx = []
            for tr in translations:
                coords.append(base.positions + np.dot(tr, base.cell))
                idx.append(np.arange(0, len(base)))
            idx = np.hstack(idx) # to track nearest neighbor if needed
            p = np.vstack(coords)
            dd = lineseg_dists(p, p1, p2)
            d_min = min(dd)
            min_distances.append(d_min)
        self.edge_min_dists = np.array(min_distances)



    def _filter_edges(self, cutoff, bottleneck_radius):
        mask = (self.edge_lengths <= cutoff) & (self.edge_min_dists >= bottleneck_radius)
        return self.edges[mask], mask

    

    def cutoff_search(self, dim, bottleneck_radius):
        
        """
        Calculates minimum value of a jump distance between sites in a mobile sublattice
        required to form a 1-3D percolating network.
        
        
        Parameters
        ----------

        dim: int
            percolation dimensionality, can be 1, 2, 3

        bottleneck_radius: float
            percolation threshold for an edge, i.e. minium distance between edge (line segment)
            between two sites in the mobile sublattice and the framework sublattice
        """

        if self.edges is None:
            self._collect_edges()

        if self.edge_min_dists is None:
            self._annotate_edges()
        
        rmin = 0.0
        rmax = self.upper_bound
        cutoff = -1.0
        while (rmax - rmin) > 0.01:
            probe = (rmin + rmax) / 2
            edges, _ = self._filter_edges(probe, bottleneck_radius)
            if len(edges) > 0:
                try:
                    pg = PeriodicGraph(edges[:, 0], edges[:,1], edges[:, 2:])
                    current_dim = pg.find_periodicity()[0]
                    if current_dim >= dim:
                        rmax = probe
                        cutoff = rmax
                    else:
                        rmin = probe
                except:
                    rmin = probe
            else:
                rmin = probe
        return cutoff
    


    def mincut_maxdim(self, bottleneck_radius):

        """Finds minimum cutoff for jump distances and maximum percolation dimensionality"""
        max_dim = 0
        dim_cutoff = 0.0
        for dim in (1, 2, 3):
            cutoff = self.cutoff_search(dim, bottleneck_radius)
            if cutoff > 0.0:
                max_dim = dim
                dim_cutoff = cutoff
        return  dim_cutoff, max_dim


                
    def percolation_threshold(self, dim, cutoff = 10.0):

        """
        Calculates maximum distance between percolating edges
        and the framework sublattice. 
        
        Parameters
        ----------

        dim: int
            percolation dimensionality, can be 1, 2, 3

        cutoff: float, 10.0 by default
            maximum allowed jump distance between sites in the mobile sublattice
        """

        if self.edges is None:
            self._collect_edges()

        if self.edge_min_dists is None:
            self._annotate_edges()
        
        rmin = 0
        rmax = cutoff
        bottleneck_radius = 0
        while (rmax - rmin) > 0.01:
            probe = (rmin + rmax) / 2
            edges, _ = self._filter_edges(cutoff, probe)
            if len(edges) > 0:
                try:
                    pg = PeriodicGraph(edges[:, 0], edges[:,1], edges[:, 2:])
                    current_dim = pg.find_periodicity()[0]
                    if current_dim >= dim:
                        rmin = probe
                        bottleneck_radius = round(rmin,4)
                    else:
                        rmax = probe
                except:
                    rmax = probe
            else:
                rmax = probe
        return bottleneck_radius
    


    def _inequivalent_edges(self, edges):
        
        decimals = int(np.ceil(-np.log10(self.symprec))) # precision
        frac_pos = self.atoms.get_scaled_positions()
        cell = self.atoms.cell

        if self.symprec:
            dataset = get_symmetry_dataset((cell, frac_pos, self.atoms.numbers),
                                            symprec=self.symprec)
            rotations = dataset['rotations']
            translations = dataset['translations']
        else:
            rotations = [np.eye(3)]
            translations = [np.zeros(3)]

        keys = []
        for _, edge in enumerate(edges):

            i, j = int(edge[0]), int(edge[1])
            offset = edge[2:]

            p_i = frac_pos[i]
            p_j = frac_pos[j] + offset
            d_ij = p_j - p_i

            orbit_keys = []
            for R, t in zip(rotations, translations):
                p_sym = (R @ p_i + t) % 1.0
                d_sym = R @ d_ij

                k1 = np.concatenate([p_sym, d_sym]).round(decimals)
                k2 = np.concatenate([(p_sym + d_sym) % 1.0, -d_sym]).round(decimals)
                
                orbit_keys.append(min(tuple(k1), tuple(k2)))
            edge_key = min(orbit_keys)
            keys.append(edge_key)
        self.edge_keys = np.array(keys)
        _, index, inverse = np.unique(keys, axis = 0, return_index=True, return_inverse=True)
        return edges[index], index, inverse
    


    def _inequivalent_edges_naive(self, edges, lengths, min_dists):

        decimals = int(np.ceil(-np.log10(self.symprec))) # precision
        atoms = self.atoms.copy()
        spglib_cell = (atoms.cell,
                       atoms.get_scaled_positions(),
                       atoms.numbers)
        equivalent_sites = get_symmetry_dataset(spglib_cell,
                                                self.symprec)['equivalent_atoms']
        
        ij = np.column_stack([
            np.take(equivalent_sites, edges[:, 0]),
            np.take(equivalent_sites, edges[:, 1]),
        ])


        ij.sort(axis = 1)
        features = np.column_stack([ij, lengths, min_dists]).round(decimals)
        _, index, inverse = np.unique(features, axis=0, return_index=True, return_inverse=True)
        return edges[index], index, inverse



    def unique_edges(self, cutoff, bottleneck_radius, method = 'naive'):

        """Find inequivalent jumps [source, target, offset_x, offset_y, offset_z]
        
        Parameters
        ----------

        cutoff: float
            maximum distance of the jump, i.e. edge length

        bottleneck_radius: float
            Minium allowed distance between the edge and the framework sublattice

        method: str, can be "naive" (default) or "symop"
            method to detect symmetrically equivalent edges.
            - "naive" uses wyckoff positions, edge length, 
              and a minimum distance between the edge and the framework 
              to find unique edges
            
            - "symop" uses symmetry operations to generate all possible edges
              and select inequivalent ones

            "naive" method is faster and yields smaller number of edges,
            but stronger depends on numerical accuracy.

            "symop" is safer but may output equivalent edges which may results
            in a non-efficient use of compute

        
        Returns
        ----------
        inequivalent jumps - array of [source, target, offset_x, offset_y, offset_z]

        """
        
        if self.edges is None:
            self._collect_edges()

        if self.edge_min_dists is None:
            self._annotate_edges()

        edges, mask = self._filter_edges(cutoff, bottleneck_radius)
        if method == 'naive':
            lengths = self.edge_lengths[mask]
            min_dists = self.edge_min_dists[mask]
            unique_edges, index, inverse = self._inequivalent_edges_naive(edges,lengths, min_dists)
        elif method == 'symop':
            unique_edges, index, inverse = self._inequivalent_edges(edges)
        else:
            raise ValueError(f'Wrong method ({method}). Can be "naive" or "symop"')
        return unique_edges
    

    def find_percolation_barriers(self, cutoff, bottleneck_radius, emins_unique, emaxs_unique, method):

        """
        Compute 1D, 2D, and 3D percolation energy barriers using precomputed minimum
        and maximum energies for unique ionic jumps forming a percolation network at
        a given jump distance cutoff and bottleneck radius.

        Parameters
        ----------

        cutoff: float
            maximum distance of the jump, i.e. edge length

        bottleneck_radius: float
            Minium allowed distance between the edge and the framework sublattice


        emins_unique: list of N x float
            minimum energies in the energy profile for unique edges
            obtained using self.unique_edges

        emaxs_unique: list of N x float
            maximum energies in the energy profile for unqiue edges
            obtained using self.unique_edges

        method: str, can be "naive" or "symop"
            method to detect symmetrically equivalent edges.
            must be consistent with the method used in self.unique_edges

        Returns
        -------
        dict with "e1d", e2d", and "e3d" percolation barriers
        
        If percolation is not possible for a given dimension, the value is `None`.

        """

        barriers = {}
        for dim in [1, 2, 3]:
            try:
                result = self._propagate_barriers(cutoff, bottleneck_radius, emins_unique, emaxs_unique, dim, method)
                barriers.update({f'e{dim}d': result[f'e{dim}d']})
            except:
                barriers.update({f'e{dim}d': None})
        return barriers
                


    def _propagate_barriers(self, cutoff, bottleneck_radius, emins_unique, emaxs_unique, dim,
                           method):
        
        """
        Compute percolation energy barrier for a provided dimensionality using precomputed minimum
        and maximum energies for unique ionic jumps forming a percolation network at
        a given jump distance cutoff and bottleneck radius.

        Parameters
        ----------

        cutoff: float
            maximum distance of the jump, i.e. edge length

        bottleneck_radius: float
            Minium allowed distance between the edge and the framework sublattice


        emins_unique: list of N x float
            minimum energies in the energy profile for unique edges
            obtained using self.unique_edges

        emaxs_unique: list of N x float
            maximum energies in the energy profile for unqiue edges
            obtained using self.unique_edges

        dim: int
            percolation dimensionality, can be 1, 2, 3

        method: str, can be "naive" or "symop"
            method to detect symmetrically equivalent edges.
            must be consistent with the method used in self.unique_edges

        """
        
        if method == 'naive':
            edges, mask = self._filter_edges(cutoff, bottleneck_radius)
            lengths = self.edge_lengths[mask]
            min_dists = self.edge_min_dists[mask]
            _, _, inverse = self._inequivalent_edges_naive(edges,
                                                               lengths,
                                                               min_dists)
        elif method == 'symop':
            edges, _ = self._filter_edges(cutoff, bottleneck_radius)
            _, _, inverse = self._inequivalent_edges(edges)
        else:
            raise ValueError(f'Wrong method ({method}). Can be "naive" or "symop"')


        emins = np.asarray(emins_unique)[inverse]
        emaxs = np.asarray(emaxs_unique)[inverse]

        upper_barrier = None
        for E in np.unique(emaxs):
            mask = emaxs <= E
            current_edges = edges[mask]

            if len(current_edges) == 0:
                continue

            pg = PeriodicGraph(current_edges[:, 0], current_edges[:,1], current_edges[:, 2:])
            current_dim = pg.find_periodicity()[0]
            if current_dim >= dim:
                upper_barrier = E
                break

        if upper_barrier is None:
            raise RuntimeError("No percolating network found")


        lower_barrier = None

        mask_upper = emaxs <= upper_barrier
        edges_upper = edges[mask_upper]
        emins_upper = emins[mask_upper]

        for E in np.unique(emins_upper):
            mask = emins_upper >= E
            current_edges = edges_upper[mask]

            if len(current_edges) == 0:
                continue

            pg = PeriodicGraph(current_edges[:, 0], current_edges[:,1], current_edges[:, 2:])
            current_dim = pg.find_periodicity()[0]
            if current_dim >= dim:
                lower_barrier = E
                break

        if lower_barrier is None:
            raise RuntimeError("Lower barrier not found")
        
        result = {
            f'e{dim}d': upper_barrier - lower_barrier,
            'emax': upper_barrier,
            'emin': lower_barrier,
            'edges': current_edges,
        }
        
        return result


import numpy as np
from ase.data import covalent_radii
from ase.neighborlist import NeighborList
from ase.calculators.calculator import Calculator, all_changes



class GMCalculator(Calculator):
    """
    Geometric median calculator
    """

    implemented_properties = ["energy", "forces", "min_distance"]
    default_parameters = {
        'rc': 10.0,
    }
    nolabel = True # check

    def __init__(self, site = None, nl = None, covalent = False, **kwargs):
        """

        Parameters
        ----------
        site: int
            indices of sites of interest in ase.Atoms
            Note: Considering source, taret notation it is the source
        
        nl: ase NeighborList, optional
            object should have method get_neighbors(site)
        
        """
        super().__init__(**kwargs)
        self.site = site
        self.nl = nl
        self.covalent = covalent

    def calculate(
        self,
        atoms = None,
        properties = None,
        system_changes = None,
    ):
        """
        Args:
            atoms (ase.Atoms): ase Atoms object
            properties (list): list of properties to calculate
            system_changes (list): monitor which properties of atoms were
                changed for new calculation. If not, the previous calculation
                results will be loaded.
        Returns:

        """
        properties = properties or ["energy"]
        system_changes = system_changes or all_changes
        super().calculate(atoms=atoms, properties=properties, system_changes=system_changes)
        
        natoms = len(self.atoms)
        rc = self.parameters.rc

        if self.covalent:
            self.rc = [covalent_radii[n] for n in self.atoms.numbers]

        if self.nl is None or 'numbers' in system_changes:
            self.nl = NeighborList(
                [rc / 2] * natoms, self_interaction=False, bothways=True, skin = 0.3
            )

        self.nl.update(self.atoms)

        energies = np.zeros(natoms)
        forces = np.zeros((natoms, 3))
        #positions = self.atoms.positions
        #cell = self.atoms.cell
        neighbors, offsets = self.nl.get_neighbors(self.site)

        r_vec = self.atoms.positions[self.site] - (atoms.positions[neighbors] + np.dot(offsets, atoms.cell))
        r_norm = np.linalg.norm(r_vec, axis = 1)
        if self.covalent:
            r_norm -= np.take(self.rc, neighbors)

        f = (1/r_norm).sum()
        energies[self.site] = f
        energy = energies.sum()
        x, y, z = r_vec[:, 0], r_vec[:, 1], r_vec[:, 2]
        dfdx = (-x/(r_norm ** 3)).sum()
        dfdy = (-y/(r_norm ** 3)).sum()
        dfdz = (-z/(r_norm ** 3)).sum()
        grad = np.array([dfdx, dfdy, dfdz])
        forces[self.site] = -grad
        self.results['energy'] = energy
        self.results['energies'] = energies
        self.results['free_energy'] = energy
        self.results['forces'] = forces
        self.results['min_distance'] = r_norm.min()


import itertools
import numpy as np
from scipy.spatial import cKDTree

from ase.data import covalent_radii
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.analysis.local_env import VoronoiNN

from ..edge import Edge 
from ..geometry import VoroSite, _geometric_median, lineseg_dists, project_point_on_edge
from ..data import ionic_radii, elneg_pauling




class EdgeFeaturizer(Edge):

    """
    Featurizes edge connecting source and target 
    positions of a mobile atom 

    Most of the features are derived from the Voronoi partitioning
    Works similarly to the pymatgen' VoronoiNN 
    but should be faster as the smaller r_cut is needed

    We use solid_angle/max(solid_angle) as the weights for each neighbor

    The following descriptors are calculated for source, target, gm, and nnp positions
    of a mobile ion, and statistics of some of these:

    CN
    volume
    area
    face_dist
    ionic_dist
    elneg
    weighted_elneg
    oxi_state
    weighted_oxi_state
    length
    cylinder_volume
    cylinder_volume^1/3
    cylinder_ionic_volume^1/3
    has_Element
    """


    def __init__(self, edge, oxi = False, use_VoronoiNN = False, bvse = False):

        """
        Parameters
        ----------

        edge: ions.geom.Edge object
            edge connecting source and target positions of a mobile atom
            
        oxi: boolean, False by defalut
            whether collect features associated with oxidation states
            Note: edge.atoms should contain 'oxi_states' array
        
        bvse: boolean, False by defalut
            whether collect features associated with BVSE parameters (alpha, d0, r_min)
            Note: edge.atoms should contain 'oxi_states' array
            Not implemented yet

        use_VoronoiNN: boolean, False by defalut
            use pymatgen's VoronoiNN instead of the custom ions.geom.utils.VoroSite
            VornoiNN is supposed to be more accurate but takes more compute time

        """

        super().__init__(edge.atoms.copy(), edge.source, edge.target, edge.offset)
        self.st = AseAtomsAdaptor.get_structure(self.atoms.copy())
        self.use_VoronoiNN = use_VoronoiNN
        self._set_covalent_radii()
        self._set_elneg()
        self.oxi = oxi
        if self.oxi:
            self._set_ionic_radii()
        # if self.bvse:
        #     collect_bvse_params(self.atoms, self.atoms.symbols[self.target], self.atoms.arrays['oxi_states'][self.source])


    def __repr__(self):
        return f"EdgeFeaturizer(source={self.source}, target={self.target}, offset={self.offset}, atoms={self.atoms})"


    def _set_ionic_radii(self):
        ri = np.array([ionic_radii[s][q] for (s, q) in zip(self.atoms.symbols, self.atoms.arrays['oxi_states'])])
        self.atoms.set_array('ri', ri)

    def _set_covalent_radii(self):
        rc = np.array([covalent_radii[n] for n in self.atoms.numbers])
        self.atoms.set_array('rc', rc)

    def _set_elneg(self):
        elneg = np.array([elneg_pauling[s] for s in self.atoms.symbols])
        self.atoms.set_array('elneg', elneg)



    def _vcutoff(self):
        """from pymatgen.analysis.local_env"""
        corners = [[1, 1, 1], [-1, 1, 1], [1, -1, 1], [1, 1, -1]]
        d_corners = [np.linalg.norm(self.atoms.cell.cartesian_positions(c)) for c in corners]
        max_cutoff = max(d_corners) + 0.01
        return max_cutoff

    def _length(self):
        return super().length
    
    def min_distance_to_edge(self):
        rest_idx = [i for i in range(len(self.atoms)) if i not in (self.source, self.target)]
        positions = self.atoms.positions[rest_idx]
        translations = np.array(list(itertools.product(
                                        [0, 1, -1], # should be [0, 1, -1, 2, -2] ideally
                                        [0, 1, -1],
                                        [0, 1, -1])))
        coords = []
        idx = []
        for tr in translations:
            coords.append(positions + np.dot(tr, self.cell))
            idx.append(rest_idx)
        ii = np.hstack(idx)
        p = np.vstack(coords)
        #self._coords = p
        dd  = lineseg_dists(p, self.p1, self.p2)
        d_min = min(dd)
        nn_id = ii[dd == dd.min()].ravel()[0]
        nearest_point = p[dd == dd.min()]
        edge_point = project_point_on_edge(nearest_point, self.p1 , self.p2)
        return d_min, nn_id, edge_point

    def _effective_CN(self, areas):
        CN = (areas.sum()**2) / ((areas**2).sum())
        return CN
    

    
    def _calc_stat(self, arr, stat):

        if stat == 'min':
            return arr.min()
        if stat =='max':
            return arr.max()
        if stat == 'range':
            return arr.max() - arr.min()
        if stat == 'std':
            return arr.std()
        if stat == 'mean':
            return arr.mean()   
    
    def _cylinder_volume_of_max_percolation_radius(self, min_dist):
        return np.pi * (min_dist**2) * self.length


    
    def _get_voro_data(self, atoms, site_id):
        if self.use_VoronoiNN:
            # shift = np.dot([0.5, 0.5, 0.5], atoms.cell) - atoms.positions[site_id]
            # atoms.positions += shift
            # atoms.wrap()
            st = AseAtomsAdaptor.get_structure(atoms)
            calc = VoronoiNN(cutoff = self._vcutoff(), weight='solid_angle')
            data_voro = calc.get_nn_info(st, site_id)
            #self.data_voro = data_voro
            data = {
                'volume': np.array([dat['poly_info']['volume'] for dat in data_voro]),
                'area': np.array([dat['poly_info']['area'] for dat in data_voro]),
                'face_dist': np.array([dat['poly_info']['face_dist'] for dat in data_voro]),
                'solid_angle': np.array([dat['poly_info']['solid_angle'] for dat in data_voro]),
                'nn_id_unitcell': np.array([dat['site_index'] for dat in data_voro]),
                'weight': np.array([dat['weight'] for dat in data_voro])
            }
        else:
            calc = VoroSite(atoms)
            data, supercell = calc.get_poly_data(site_id)
            data['weight'] = data['solid_angle'] / data['solid_angle'].max()
            
        data.update({'rc': atoms.arrays['rc'][data['nn_id_unitcell']]})
        data.update({'elneg': atoms.arrays['elneg'][data['nn_id_unitcell']]})
        data.update({'elneg_weighted': data['weight'] * atoms.arrays['elneg'][data['nn_id_unitcell']]})
        if self.oxi:
            data.update({'ri': atoms.arrays['ri'][data['nn_id_unitcell']]})
            data.update({'oxi_state': atoms.arrays['oxi_states'][data['nn_id_unitcell']]})
            data.update({'oxi_state_weighted': data['weight'] * atoms.arrays['oxi_states'][data['nn_id_unitcell']]})
            data.update({'ionic_dist': 2 * data['face_dist'] - data['ri']})
        return data
    
    def _get_voro_features(self):

        #source
        data_source = self._get_voro_data(self.atoms, self.source)
        #target
        data_target = self._get_voro_data(self.atoms, self.source)

        # nnp
        d_min, nn_id, pos = self.min_distance_to_edge()
        atoms = self.atoms.copy()
        site_id = min(self.source, self.target)
        atoms.positions[site_id] = pos
        atoms = atoms[[i for i in range(len(atoms)) if i != max(self.source, self.target)]]
        data_nnp = self._get_voro_data(atoms, site_id)

        # gm
        CN = len(data_nnp['nn_id_unitcell'])
        atoms = self.atoms.copy()
        site_id = min(self.source, self.target)
        rest_idx = [i for i in range(len(atoms)) if i not in (self.source, self.target)]
        atoms = atoms[[i for i in range(len(atoms)) if i != max(self.source, self.target)]]
        positions = self.atoms.positions
        translations = np.array(list(itertools.product(
                                        [0, 1, -1], # should be [0, 1, -1, 2, -2] ideally
                                        [0, 1, -1],
                                        [0, 1, -1])))
        coords = []
        idx = []
        for tr in translations:
            coords.append(positions[rest_idx] + np.dot(tr, self.cell))
            idx.append(rest_idx)

        ii = np.hstack(idx)
        p = np.vstack(coords)
        tree = cKDTree(p)
        dd, ii = tree.query(pos, distance_upper_bound=10.0, k = CN) 
        gm = _geometric_median(p[ii], pos) # the initial guess is from nnp
        atoms.positions[site_id] = gm
        data_gm = self._get_voro_data(atoms, site_id)
        return data_source, data_target, data_nnp, data_gm
     

    def _is_element_in_structure(self, possible_elements = ['O', 'F', 'P', 'S', 'B', 'I', 'F', 'Cl', 'N']):
        data = {}
        for element in possible_elements:
            data.update({f'has_{element}': int(element in self.atoms.symbols)})
        return data

    

    def featurize(self, stats = ['min', 'max', 'mean', 'range']):

        """
        Featurize edge

        Parameters
        ----------

        stats: list of str, ['min', 'max', 'mean', 'range'] by default
            which feature statistics to calculate
            possible stats are 'min', 'max', 'mean', 'range', 'std'
        
        Returns
        ----------
        features: dict, {feature_label: feature, ...}
            calculated features
        """

        features = {}
        features.update({'length': self._length()})
        d_min, nn_id, pos = self.min_distance_to_edge()
        features.update({'min_dist_to_edge': d_min})

        data_source, data_target, data_nnp, data_gm = self._get_voro_features()

        percolation_radii = 2 * min(
                                data_source['face_dist'].min(),
                                data_target['face_dist'].min(),
                                data_nnp['face_dist'].min(),
                                data_gm['face_dist'].min()
                                )
        features.update({'cylinder_volume': self._cylinder_volume_of_max_percolation_radius(percolation_radii)})
        features.update({'cylinder_volume^1/3': features['cylinder_volume'] ** 1/3}) 
        if self.oxi:
            percolation_ionic_radii =  min(
                                data_source['face_dist'].min(),
                                data_target['face_dist'].min(),
                                data_nnp['face_dist'].min(),
                                data_gm['face_dist'].min()
                                )
        if self.oxi:
            features.update({'cylinder_ionic_volume': self._cylinder_volume_of_max_percolation_radius(percolation_ionic_radii)})
            features.update({'cylinder_ionic_volume^1/3': features['cylinder_ionic_volume'] ** 1/3}) 

        CN_source = self._effective_CN(data_source['area'])
        CN_target = self._effective_CN(data_target['area'])
        CN_nnp = self._effective_CN(data_nnp['area'])
        CN_gm = self._effective_CN(data_gm['area'])
        arr = np.array([CN_source, CN_target, CN_nnp, CN_gm])
        for stat in stats:
            features.update({f'{stat}_CN': self._calc_stat(arr, stat = stat)})


        total_volume_source = data_source['volume'].sum()
        total_volume_target = data_target['volume'].sum()
        total_volume_nnp = data_nnp['volume'].sum()
        total_volume_gm = data_gm['volume'].sum()
        arr = np.array([total_volume_source, total_volume_target, total_volume_nnp, total_volume_gm])
        for stat in stats:
            features.update({f'{stat}_volume': self._calc_stat(arr, stat = stat)})



        total_area_source = data_source['area'].sum()
        total_area_target = data_target['area'].sum()
        total_area_nnp = data_nnp['area'].sum()
        total_area_gm = data_gm['area'].sum()
        arr = np.array([total_area_source, total_area_target, total_area_nnp, total_area_gm])
        for stat in stats:
            features.update({f'{stat}_area': self._calc_stat(arr, stat = stat)})



        for property in data_source.keys():
            if property in ['weight', 'nn_id', 'nn_id_unitcell']:
                continue
            for stat1 in stats:
                property_source = self._calc_stat(data_source[property], stat = stat1)
                property_target = self._calc_stat(data_target[property], stat = stat1)
                property_nnp = self._calc_stat(data_nnp[property], stat = stat1)
                property_gm = self._calc_stat(data_gm[property], stat = stat1)
                arr = np.array([property_source, property_target, property_nnp, property_gm])
                for stat2 in stats:
                    features.update({f'{stat2}_{stat1}_{property}': self._calc_stat(arr, stat = stat2)})
        features.update(self._is_element_in_structure())
        return features